import json
import tempfile
from pathlib import Path

from cleez import Command
from cleez.command import Argument

from ..client import get_client
from ._common import NUA_ENV, NUA_HOST, get_config, sh, ssh

client = get_client()


# Hardcoded for now
class DeployCommand(Command):
    """Deploy app."""

    name = "deploy"

    arguments = [
        Argument("domain", help="Domain to deploy to."),
    ]

    def run(self, domain=None):
        """Deploy all apps."""
        config = get_config()
        host = NUA_HOST

        app_id = config["metadata"]["id"]
        deploy_config = {
            "site": [
                {
                    "image": app_id,
                    "domain": domain,
                }
            ],
        }

        temp_dir = tempfile.mkdtemp(prefix=f"nua-deploy-{app_id}-", dir="/tmp")  # noqa
        config_file = Path(temp_dir) / "deploy.json"
        Path(config_file).write_text(json.dumps(deploy_config, indent=2))

        sh(f"rsync -az --mkpath {config_file} root@{host}:{config_file}")
        ssh(f"{NUA_ENV}/bin/nua-orchestrator deploy {config_file}", host)



from __future__ import annotations

from pathlib import Path

import tomli

# from .version import get_version


# def _version_callback(value: bool) -> None:
#     if value:
#         print_version()
#         raise typer.Exit(0)
#
#
# OPTS = {
#     "version": typer.Option(
#         None,
#         "--version",
#         "-V",
#         help="Show Nua version and exit.",
#         callback=_version_callback,
#         is_eager=True,
#     ),
#     "verbose": typer.Option(
#         0, "--verbose", "-v", help="Show more informations, until -vvv.", count=True
#     ),
#     "color": typer.Option(True, "--color/--no-color", help="Colorize messages."),
# }
#
#
# def print_version() -> None:
#     typer.echo(f"Nua CLI version: {get_version()}")


def get_current_app_id() -> str:
    """Get the current app id (if possible)."""
    try:
        config = get_current_app_config()
        return config["metadata"]["id"]
    except ValueError:
        return ""


def get_current_app_config() -> dict:
    config_file = Path("nua-config.toml")
    if not config_file.exists():
        config_file = Path("nua/nua-config.toml")
    if not config_file.exists():
        raise ValueError("No nua-config.toml found.")

    config_file = Path("nua/nua-config.toml")
    config_data = config_file.read_text()
    return tomli.loads(config_data)



from cleez.colors import bold, red
from cleez.command import Argument, Command

from ..client import get_client
from ..common import get_current_app_id

client = get_client()


class EnvCommand(Command):
    """Manage an app's env variables."""

    name = "env"
    hide_from_help = True

    def run(self):
        self.cli.print_help()


class EnvShowCommand(Command):
    """Show application env variables."""

    name = "env show"

    arguments = [
        Argument("app_id", nargs="?", help="Application ID"),
    ]

    def run(self, app_id: str = ""):
        if not app_id:
            app_id = get_current_app_id()
        app_info = client.get_app_info(app_id)

        print(bold("Environment variables from the configuration:"))
        for k, v in sorted(app_info["site_config"]["env"].items()):
            print(f"{k}={repr(v)}")

        container_id = app_info["site_config"]["container_id"]
        try:
            container_info = client.get_container_info(container_id)
        except ValueError:
            print(red("Container not found - app is probably not running"))
            return

        print()
        print(bold("Environment variables from container:"))
        container_env = container_info[0]["Config"]["Env"]
        for k in sorted(container_env):
            print(k)


class EnvSetCommand(Command):
    """Show application env variables."""

    name = "env set"

    def run(self, app_id: str = ""):
        print(red("Not implemented yet"))



from cleez import Command
from cleez.colors import green, red, yellow

from ..client import get_client

client = get_client()


class AppsCommand(Command):
    """List applications."""

    name = "apps"

    def run(self):
        result = client.call("list")
        for instance in result:
            app_id = instance["app_id"]
            domain = instance["site_config"]["domain"]
            container_id = instance["site_config"]["container_id"]

            try:
                container_info = client.get_container_info(container_id)
            except ValueError:
                print(red(f"{app_id} @ {domain} - container not found"))
                continue

            status = container_info[0]["State"]["Status"]

            match status:
                case "running":
                    color = green
                case "exited":
                    color = red
                case _:
                    color = yellow

            print(color(f"{app_id} @ {domain} ({status})"))



import os
import shlex
import subprocess
from pathlib import Path

import tomli
from cleez.colors import dim

# Hardcoded for now
NUA_ENV = "/home/nua/env"
ORCH_PATH = "/home/nua/env/bin/nua-orchestrator"
NUA_HOST = os.environ["NUA_HOST"]


#
# helpers
#
def sh(cmd: str, cwd: str = "."):
    """Run a shell command."""
    print(dim(f'Running "{cmd}" locally in "{cwd}"...'))
    args = shlex.split(cmd)
    try:
        subprocess.run(args, check=True)
    except subprocess.CalledProcessError as e:
        print(f"Command failed: {e.cmd}")


def ssh(cmd: str, host: str):
    """Run a ssh command."""
    print(dim(f'Running "{cmd}" on server...'))
    args = shlex.split(cmd)
    cmd = ["ssh", f"nua@{host}", f"{shlex.join(args)}"]  # type: ignore
    try:
        subprocess.run(cmd, check=True)
    except subprocess.CalledProcessError as e:
        print(f"Command failed: {e.cmd}")


def get_config() -> dict:
    config_file = Path("nua-config.toml")
    if not config_file.exists():
        config_file = Path("nua/nua-config.toml")
    config_data = config_file.read_text()
    return tomli.loads(config_data)



from pprint import pp

from cleez import Command, CommandError
from cleez.command import Argument

from ..client import get_client
from ..common import get_current_app_id

client = get_client()


class ConfigCommand(Command):
    """Show/edit application config."""

    name = "config"
    hide_from_help = True

    def run(self):
        self.cli.print_help()


class ShowCommand(Command):
    """Show application config."""

    name = "config show"

    arguments = [
        Argument("app_id", nargs="?", help="Id of the application."),
    ]

    def run(self, app_id: str):
        result = client.call("list")

        if not app_id:
            app_id = get_current_app_id()
        if not app_id:
            raise CommandError("No app_id specified")

        for instance in result:
            if instance["app_id"] == app_id:
                pp(instance)
                break
        else:
            raise CommandError(f"App {app_id} not found")



from operator import itemgetter
from pprint import pp

from cleez import BadArgumentError
from cleez.command import Argument, Command

from ..client import get_client

client = get_client()


class ServerCommand(Command):
    """Manage the Nua server."""

    name = "server"
    hide_from_help = True

    def run(self):
        self.cli.print_help()


class LogsCommand(Command):
    """Show server logs."""

    name = "server logs"

    arguments = [
        Argument("service", help="Service to show logs for"),
    ]

    def run(self, service: str):
        if not service:
            print("Service must be one of: nua, letsencrypt, nginx")

        match service:
            case "nua":
                print("Showing Nua logs [TODO]")
            case "letsencrypt":
                result = client.ssh("cat log/letsencrypt/letsencrypt.log")
                print(result.stdout)
            case "nginx":
                print("Showing Nginx logs [TODO]")
            case _:
                raise BadArgumentError(
                    "Service must be one of: nua, letsencrypt, nginx"
                )


class StatusCommand(Command):
    """Show Nua status."""

    name = "server status"

    def run(self):
        result = client.call("status")

        print(f"Nua version: {result['version']}")

        registries = result["registries"]
        print("Configured registries:")
        for reg in sorted(registries, key=itemgetter("priority")):
            msg = (
                f'  priority: {reg["priority"]:>2}   '
                f'format: {reg["format"]:<16}   '
                f'url: {reg["url"]}'
            )
            print(msg)


class SettingsCommand(Command):
    """Show server settings."""

    name = "server settings"

    def run(self):
        result = client.call("settings")
        pp(result)


class PsCommand(Command):
    """List all server processes."""

    name = "server ps"

    def run(self):
        result = client.ssh("ps -aux")
        print(result.stdout)


class UptimeCommand(Command):
    """Show server uptime."""

    name = "server uptime"

    def run(self):
        result = client.ssh("uptime")
        print(result.stdout)


class CleanupCommand(Command):
    """Cleanup server (remove inactive docker images and containers)."""

    name = "server cleanup"

    # TODO: ask for confirmation

    def run(self):
        result = client.ssh("docker system prune -af")
        result = client.ssh("docker volume prune -f")
        print(result.stdout)



from cleez.colors import red
from cleez.command import Argument, Command

from ..client import get_client
from ._common import ORCH_PATH

client = get_client()


class StartCommand(Command):
    """Start an application."""

    name = "start"

    arguments = [
        Argument("app_id", help="Application ID"),
        Argument("domain", help="Domain name"),
    ]

    def run(self, app_id: str, domain: str):
        result = client.ssh(f"{ORCH_PATH} start -d {domain}")
        print(result.stdout)


class StopCommand(Command):
    """Stop an application."""

    # TODO: add --all option

    name = "stop"

    arguments = [
        Argument("app_id", help="Application ID"),
        Argument("domain", help="Domain name"),
    ]

    def run(self, app_id: str, domain: str):
        result = client.ssh(f"{ORCH_PATH} stop -d {domain}")
        print(result.stdout)


class RestartCommand(Command):
    """Restart an application."""

    name = "restart"

    arguments = [
        Argument("app_id", help="Application ID"),
        Argument("domain", help="Domain name"),
    ]

    def run(self, app_id: str, domain: str):
        result = client.ssh(f"{ORCH_PATH} restart -d {domain}")
        print(result.stdout)


class DestroyCommand(Command):
    """Destroy an application."""

    name = "destroy"

    arguments = [
        Argument("app_id", help="Application ID"),
        Argument("domain", help="Domain name"),
    ]

    # TODO: add confirmation step

    def run(self, app_id: str, domain: str):
        result = client.ssh(f"{ORCH_PATH} remove -d {domain}")
        print(result.stdout)


class UpdateCommand(Command):
    """Update an application."""

    name = "update"

    def run(self):
        print(red("Not implemented yet"))



import tempfile

from cleez import Command
from cleez.colors import bold

from ..client import get_client
from ._common import NUA_ENV, NUA_HOST, get_config, sh, ssh

client = get_client()

# Hardcoded for now
EXCLUDES = [
    ".git",
    ".env",
    ".venv",
    ".cache",
    ".mypy_cache",
    ".pytest_cache",
    ".ruff_cache",
    ".nox",
    ".tox",
    ".coverage",
    ".idea",
    "dist",
    "build",
    "data",
    "doc",
    "vite/node_modules/",
    "front-old",
    "tmp",
    "sandbox",
]


class BuildCommand(Command):
    """Build app but don't deploy it."""

    name = "build"

    def run(self):
        host = NUA_HOST
        config = get_config()
        app_id = config["metadata"]["id"]

        build_dir = tempfile.mkdtemp(prefix=f"nua-build-{app_id}-", dir="/tmp")  # noqa

        print(bold(f"Building app {app_id}..."))

        excludes_args = " ".join([f"--exclude={e}" for e in EXCLUDES])
        ssh(f"mkdir -p {build_dir}", host=host)
        sh(f"rsync -e ssh --delete-after -az {excludes_args} ./ nua@{host}:{build_dir}")

        ssh(f"{NUA_ENV}/bin/nua-build -vvv {build_dir}", host)

        print()


# @cli.command()
#
#     verbosity = 1 + verbose - quiet
#     verbosity_flags = verbosity * ["-v"]
#
#     if experimental:
#         typer.secho(
#             "Using experimental builder (from nua-dev).", fg=typer.colors.YELLOW
#         )
#         subprocess.run(["nua-dev", "build", path])
#     else:
#         subprocess.run(["nua-build"] + verbosity_flags + [path])
#



from cleez.command import Argument, Command

from nua_cli.client import get_client
from nua_cli.common import get_current_app_id

client = get_client()


class LogsCommand(Command):
    """Show application logs."""

    name = "logs"

    arguments = [
        Argument("app_id", nargs="?", help="Application ID"),
    ]

    def run(self, app_id: str = ""):
        # Quick & dirty implementation that calls docker directly.
        if not app_id:
            app_id = get_current_app_id()
        app_info = client.get_app_info(app_id)
        container_id = app_info["site_config"]["container_id"]
        result = client.ssh(f"docker logs {container_id}")
        # Note: 'docker logs' returns data to both stderr and stdout.
        # We need to merge the two streams, on find another way
        print(result.stderr)



import json
import os
import sys
from io import StringIO

from cleez.colors import red
from fabric import Connection

# Hardcoded for now
NUA_CMD = "./env/bin/nua-orchestrator"


class Client:
    connection: Connection

    def __init__(self, host: str = "", user: str = ""):
        if not host:
            host = os.environ.get("NUA_HOST", "localhost")
        if not user:
            user = os.environ.get("NUA_USER", "nua")

        self.host = host
        self.user = user
        self.connection = Connection(self.host, self.user)

    #
    # Low level API
    #
    def call_raw(self, method: str, **kw):
        args = StringIO(json.dumps(kw))
        cmd = f"{NUA_CMD} rpc --raw {method}"
        r = self.connection.run(cmd, hide=True, in_stream=args)
        return r.stdout

    def call(self, method: str, **kw):
        args = StringIO(json.dumps(kw))
        cmd = f"{NUA_CMD} rpc {method}"
        r = self.connection.run(cmd, hide=True, in_stream=args)
        try:
            return json.loads(r.stdout)
        except json.JSONDecodeError:
            print(red(f"Invalid response from server:\n{r.stdout}"))
            sys.exit(1)

    def ssh(self, command: str):
        return self.connection.run(command, hide=True)

    #
    # Higher level API
    #
    def get_app_info(self, app_id: str) -> dict:
        result = self.call("list")

        app_info = None
        for instance in result:
            if instance["app_id"] == app_id:
                app_info = instance
                return app_info

        raise ValueError(f"App {app_id} not found")

    def get_container_info(self, container_id: str) -> dict:
        try:
            result = self.ssh(f"docker inspect {container_id}").stdout
            return json.loads(result)
        except:  # noqa
            raise ValueError(f"Container {container_id} not found")


_CLIENT = None


def get_client(host: str = "", user: str = ""):
    global _CLIENT

    if not _CLIENT:
        _CLIENT = Client(host, user)
    return _CLIENT



from cleez import Command
from cleez.colors import red

from ..client import get_client

client = get_client()


class BackupCommand(Command):
    """Backup a deployed application."""

    name = "backup"

    def run(self):
        result = client.call_raw("backup")
        print(result)


class RestoreCommand(Command):
    """Restore backup data of a deployed application."""

    name = "restore"

    def run(self):
        print(red("Not implemented yet"))



"""
TODO:

[x] = done
[ ] = not done

[x] apps              List apps
[x] build             Build app but don't deploy it
[x] backup            Backup/restore app data
[x] config            Show/manage config for current app
[x] deploy            Deploy app
[x] destroy           Destroy app
[x] help              Display help
[ ] init              Create a new app
[x] logs              Tail running logs
[x] restart           Restart an app
[ ] run               Run a command in the app's environment
[ ] scale             Scale processes
[x] server            Manage the Nua server
[x] start             Start an app
[x] status            Show app status
[x] stop              Stop an app
[ ] update            Update an app
[x] version           Show Nua version
"""
from __future__ import annotations

import importlib.metadata

import snoop
from cleez import CLI
from cleez.actions import VERSION

snoop.install()


def main():
    cli = get_cli()
    cli.run()


def get_cli():
    cli = CLI("nua", version=importlib.metadata.version("nua.cli"))
    cli.add_option(
        "-V",
        "--version",
        action=VERSION,
        version=cli.version,
        default=False,
        help="Show version and exit",
    )
    cli.add_option(
        "-d", "--debug", default=False, action="store_true", help="Enable debug mode"
    )
    cli.add_option(
        "-v", "--verbose", default=False, action="store_true", help="Increase verbosity"
    )
    cli.scan("nua_cli.commands")
    return cli


if __name__ == "__main__":
    main()



from cleez import Command

from nua_cli.client import get_client
from nua_cli.version import get_version

client = get_client()


class HelpCommand(Command):
    """Show help."""

    name = "help"

    def run(self):
        self.cli.print_help()


class VersionCommand:
    """Show Nua version."""

    name = "version"

    def run(self):
        print(f"Nua CLI version: {get_version()}")
        status = client.call("status")
        print(f"Nua Server version: {status['version']}")
